# -*- coding: iso-8859-1 -*-
"""
Dieses Modul implementiert ein normales (nicht-hidden) Markov-Modell

Vgl. Skript, "Markov-Ketten"
"""

from __future__ import generators

import bisect, random, operator

# Python 2.2 compatibility
try:
	enumerate(range(2))
except NameError:
	def enumerate(seq):
		ind = 0
		for el in seq:
			yield ind, el
			ind += 1
	def sum(seq):
		return reduce(operator.add, seq, 0)


class Error(Exception):
	pass


class ProcessAbsorbed(Error):
	pass


def normalize(seq):
	"""changes seq in place such that the sum of its elements is one.
	"""
	total = float(sum(seq))
	for i in range(len(seq)):
		seq[i] = seq[i]/total


class DistFunc:
	"""This class represents a distribution function generated from
	a distribution.  Its draw method will return items distributed
	according to this distribution.
	"""
	def __init__(self, dist):
		total = float(sum(dist))
		self.dist = [v/total for v in dist]
		self.df = []
		last = 0
		for index, val in enumerate(self.dist):
			if val>1e-10:
				last += val
				self.df.append((last, index))

	def draw(self):
		return self.df[bisect.bisect(
				self.df, (random.random(),0))][1]

	def sampleRandomly(self, samplen):
		for i in xrange(samplen):
			yield self.draw()
	
	def getDist(self):
		"""returns the distribution for the distribution function.
		"""
		return self.dist


def _addIndices(seq):
	"""returns a list of tuples (element, index) for element in seq.
	"""
	return [(seq[i], i) for i in range(len(seq))]


class MarkovModel:
	"""This markov model models the transition distributions as a
	list of DFs, each of which is again a list.  To support "named"
	states, we keep a dictionary that maps state names to state
	numbers.  The reverse is done through a list.

	To instantiate a MarkovModel, pass it a string containing a valid
	python dictionary with all states as keys and dictionaries 
	state-name:transition-probability as values.  These need not
	be normalized, but transistion probabilities must be nonzero.  
	Specify absorbing states with empty dictionaries.

	To run a Markov model, you first need to set it up, either by
	setting a state or by calling a setUpXXX function.
	"""
	def __init__(self, description):
		self.nameToNum = {}
		self.numToName = []
		self._buildTf(description)
		self.state = 0

	def _getStateNum(self, stateName):
		"""returns a state number for stateName.
		"""
		return self.nameToNum[stateName]

	def _getStateName(self, stateNum):
		"""returns a state number for stateName.
		"""
		return self.numToName[stateNum]

	def _buildStateMapping(self, stateDict):
		self.numToName = dict(enumerate(stateDict.keys()))
		self.nameToNum = dict(zip(self.numToName.values(), 
			self.numToName.keys()))

	def _computeDf(self):
		"""computes distribution functions for the transitions from
		the possibly non-normalized distributions in the transition
		function.  This sets a df attribute that must be deleted
		when you change the Tf at runtime.
		"""
		self.df = []
		for dist in self.tf:
			if dist is None:
				self.df.append(None)
			else:
				self.df.append(DistFunc(dist))
		
	def _buildTf(self, description):
		"""evals description (has to be python source for a dict of
		dicts, see above) and builds a matrix representation of 
		the transition function
		"""
		stateDict = eval(description)
		self.numStates = len(stateDict)
		self._buildStateMapping(stateDict)
		self.tf = [None]*self.numStates
		for stateName, transitions in stateDict.iteritems():
			if transitions:
				thisLine = self.tf[self._getStateNum(stateName)
					] = [0]*self.numStates
				for targetName, prob in transitions.iteritems():
					thisLine[self._getStateNum(targetName)] = prob
				normalize(thisLine)

	def setState(self, stateName):
		"""sets the current state to stateName.
		"""
		self.state = self._getStateNum(stateName)
		self.stateSeq = [self.state]

	def getState(self):
		"""returns the name of the current state.
		"""
		return self._getStateName(self.state)

	def setUpUniform(self):
		"""resets the MM to start with a state drawn uniformly from all
		known states.
		"""
		self.state = random.choice(xrange(self.numStates))
		self.stateSeq = [self.state]

	def setUpInitDf(self):
		"""resets the MM to start with a state drawn according to
		a distribution in the initDf attribute, which as to be
		set in some derived class.
		"""
		if not hasattr(self, "initDf"):
			raise Error, "Your %s object doesn't provide a initDf attribute"%(
				self.__class__)
		self.state = self.initDf.draw()
		self.stateSeq = [self.state]

	def driveOn(self):
		"""drives the Markov model one step further and returns the 
		state name of the state the model is in afterwards.
		"""
		if not hasattr(self, "df"):
			self._computeDf()
		try:
			self.state = self.df[self.state].draw()
		except AttributeError:
			raise ProcessAbsorbed, "Process absorbed after %d steps"%len(
				self.stateSeq)
		self.stateSeq.append(self.state)
		return self._getStateName(self.state)

	def getAPath(self, numSteps):
		"""returns a random path through the markov model of length
		numSteps.
		"""
		try:
			for i in range(numSteps-1):
				self.driveOn()
		except ProcessAbsorbed:
			pass
		return tuple(map(self._getStateName, self.stateSeq))

	def _computePathProb(self, numPath):
		"""returns the probability of the sequence of state indices numPath.
		"""
		try:
			return reduce(operator.mul,
				[self.tf[f][t] for f, t in zip(numPath[:-1], numPath[1:])])
		except TypeError:
			raise Error, "Impossible path"

	def computePathProb(self, path):
		"""returns the probability of the state sequence (given in Names)
		of path.
		"""
		try:
			return self._computePathProb(map(self._getStateNum, path))
		except KeyError:
			raise Error, "Invalid state name in path"


def runSingle(model, numSteps):
	model.setUpUniform()
	print "\n".join(map(str, model.getAPath(numSteps)))

def runMulti(model, numSteps, numMulti):
	for i in xrange(numMulti):
		model.setUpUniform()
		print "\n".join(map(str, model.getAPath(numSteps)))

def computeLExpect(model, numExper):
	curSum = curExper = 0.
	for i in range(numExper):
		m.setUpUniform()
		path = m.getAPath(1000)
		curSum += len(path)
		curExper += 1
		print "Letztes Ergebnis: %2d, Versuche: %4d, Mittel: %f"%(
			len(path), curExper, curSum/curExper)
	return curSum/curExper

def computeLDistri(model, numExper):
	distri = {}
	for i in range(numExper):
		m.setUpUniform()
		path = m.getAPath(1000)
		distri[path] = distri.get(path, 0)+1
	return distri

if __name__=="__main__":
	import sys
#	for val in DistFunc([1,1,1]).sampleRandomly(1000):
#		print val
#	sys.exit()
	m = MarkovModel(open(sys.argv[1]).read())
	computeLExpect(m, 20000)
